// ********************************************************************
// mifwds
//
// Forwards additive implementation for MI measurement
//
//  Author: N Dowson
//    Date: 24 Mar 06, REvised to correct off by 1 error for even sized templates
// Version: 0.6, 
// ********************************************************************

//Std c++
#include <iostream>

#include "mex.h"
#define cimg_plugin  "cimgmatlabNick.h"
#include <CImg-1-17/CImg.h>

using namespace cimg_library;

//********************************************************************
void mexErr(char* err, int nlhs, mxArray *plhs[] )
{
  std::cout << "\nAFFREGION Error: " << err << "\n" << " "
	    << "\n\nAborting...\n";
  //Create some output results
  for( int i=0; i<nlhs; i++ )
    plhs[i] = mxCreateNumericMatrix( 1, 1, mxSINGLE_CLASS, mxREAL);
}

//********************************************************************
void mexFunction(int nlhs, mxArray *plhs[], int nrhs,
                 const mxArray *prhs[])
{
  //==================================================================
  //Initialisation
  using namespace std;
  
  //Check there are 3-4 inputs
  if( nrhs!=3 ) 
    { mexErr( "3 inputs required.", nlhs, plhs ); return; }
 
  //Check there are 1-4 outputs
  if( (nlhs>1) )
    { mexErr( "1 output required", nlhs, plhs ); return; }
  
  //Read the inputs
  CImg<double> ff2tmpSz( prhs[0] ); //Template image Size
  CImg<double> ff2ref ( prhs[1], false, mxGetClassID(prhs[1]) == mxDOUBLE_CLASS  ); //Reference image
  CImg<double> ff2warp ( prhs[2] ); //The Affine warp

  CImg<double> of2region( (int) ff2tmpSz[0], (int) ff2tmpSz[1], 1,1, 0.0 );
  double *fpfPixel = of2region.ptr();

  int fiyA = (int) (-(ff2tmpSz[1]-1)/2);
  int fiyE = fiyA+(int)ff2tmpSz[1];
  int fixA = (int) (-(ff2tmpSz[0]-1)/2);
  int fixE = fixA+(int)ff2tmpSz[0];

  for( int fiy=fiyA; fiy<fiyE; fiy++ )
    for( int fix=fixA; fix<fixE; fix++ )
      {
	double ffxw = (1+ff2warp[0])*fix +    ff2warp[2] *fiy + ff2warp[4];
	double ffyw =    ff2warp[1] *fix + (1+ff2warp[3])*fiy + ff2warp[5];
	(*fpfPixel) = ff2ref.linear_pix2d(static_cast<float>(ffxw),
    static_cast<float>(ffyw) );
  //mexPrintf("%d %d  %f %f\n", fix,fiy,ffxw,ffyw);
	//of2region(fix-fixA,fiy-fiyA) = ff2ref.linear_pix2d(ffxw,ffyw);

	fpfPixel++;
      }

  plhs[0] = of2region.toMatlab();
  return;
}



