% [P3,A3,A1,A2] = overPoly( P1, P2 )
%
% Finds the polygon formed by the intersection of two simple clockwise
% polygons and its area

function [ioP1,oA,oA1,oA2] = overPoly(ioP1, iP2)

  oA1 = areaPoly(ioP1);
  oA2 = areaPoly( iP2);
  
  %Catch clockwise polygons
  if oA1<0
    disp('overPoly warning: iP1 clockwise - flipping direction.');
    ioP1 = fliplr(ioP1); oA1 = -oA1;
  end
  if oA2<0
    disp('overPoly warning: iP2 clockwise - flipping direction.');
    iP2 = fliplr(iP2); oA2 = -oA2;
  end
  
  %Slice iP1 up with every line defined by iP2
  nLine = size(iP2,2);
  iP2 = [ iP2 iP2(:,1) ];
  for cLine = 1:nLine
    ioP1 = slice( ioP1, iP2(:,cLine+[0:1]) );
  end
  
  oA = areaPoly(ioP1);
    
%---------------------------------------------------------------------------
%Calculates the area of a non-self-intersecting polygon specifed as a
%string of 2D points
function oArea = areaPoly( iPoly )
  if prod(size(iPoly))==0
    oArea = 0;
  else
    iPoly = [ iPoly iPoly(:,1) ];
    oArea = 0.5 * iPoly(1,:) * ...
      ([iPoly(2,2:end) iPoly(2,1)] - [iPoly(2,end) iPoly(2,1:end-1)])';
  end
   
%---------------------------------------------------------------------------
%Checks if p is within polygon A
% yesno =0 for on line, <0 for below, >0 for above.
function yesno = isWithin(A,p)
  Nlines = size(A,2);
  A = [A A(:,1)];

  %Start by assuming point is within polygon
  yesno = 1;
  for cLine = [1:Nlines]
    if ~isBelow( p, A(:,cLine), A(:,cLine+1) )
      yesno = 0; %If point is ever above line then it not within the polygon
      break; %So stop checking
    end
  end
  
%---------------------------------------------------------------------------
%Checks if p is above or below line ab. 
% yesno =0 for on line, <0 for below, >0 for above.
function yesno = isBelow(a,b,p)
  yesno = (a(1)-p(1))*(b(2)-p(2)) - (a(2)-p(2))*(b(1)-p(1));

%---------------------------------------------------------------------------
%Slice an existing convex polygon with a line. All points above ab are
%deleted.
function oPoly = slice( iPoly, iLine )

  if prod(size(iPoly))==0
    oPoly=[]; 
    return;
  end
  
  oPoly = [iPoly iPoly(:,1)];
  
  %Check Check for intersection of each polygon line
  c=0;
  while c < size(oPoly,2)-1
    c = c+1;
    p1 = oPoly(:,c);
    p2 = oPoly(:,c+1);
    p0 = intersect( p1, p2, iLine(:,1), iLine(:,2) );
    %Check in case no intersection ever occurs
    if prod(size(p0))~=0
      %If p0 is betw p1 and p2
      p2dotp0 = ((p2-p1)'*(p0-p1)) / ((p2-p1)'*(p2-p1));
      if p2dotp0 > 0 & p2dotp0 < 1
        %Insert point
        oPoly = [oPoly(:,1:c) p0 oPoly(:,c+1:end)];
        c = c+1;
      end
    end
  end
  
  %Delete all points above line
  c=0;
  while c < size(oPoly,2)-1
    c=c+1;
    %Numerical errors mean we can't use 0
    if -1e-4 > isBelow(oPoly(:,c),iLine(:,1),iLine(:,2)) 
      oPoly(:,c) = [];
      c=c-1;
    end
  end
  oPoly(:,end) = []; 
  
  a=0;
  
%---------------------------------------------------------------------------
%Finds the point of intersection between two lines, each specified by two
%points
function p = intersect(a,b,c,d)
  A = [[b-a] -[d-c]];
  B = [c-a];
  %Check if lines are parallel
  if 1e-4 > rcond(A) %0 == det(A)
    p = [];
  else
    S = A\B;
    p = [b-a]*S(1)+a;
  end
  