// ********************************************************************
// ssdfwds
//
// Forwards additive implementation for MI measurement
//
//  Author: N Dowson
//    Date: 12 Mar 06
// Version: 0.5
// ********************************************************************

//Std c++
#include <iostream>
#include <fstream>

//#include "mex.h"
//#define cimg_plugin  "cimgmatlabplugin.h"
#include <CImg/CImg.h>

using namespace cimg_library;

//********************************************************************
CImg<float> jointHist( CImg<float>& if2a, CImg<float>& if2b, 
			int amax=0, int bmax=0 );
CImg<float> jointHist( CImg<float>& if2a, CImg<float>& if2b,
		       CImg<float>& if2val,
		       int amax=0, int bmax=0  );
CImg<float> sumDim( CImg<float>& if2img, int iiDim=1 );
float entropy( CImg<float>& if2hist );
CImg<float> region( CImg<float>& if2tmp, CImg<float>& if2ref,  
		    CImg<float>& if2warp );
CImgl<float> imgXY( CImg<float>& if2tmp );
void outputImg( CImg<float> &if2a, char* iName );
CImgl<float> ImgListMul( CImgl<float> &ilf2a, int iima, int iina,
			 CImgl<float> &ilf2b, int iimb, int iinb );
CImgl<float> getdWdV( CImgl<float> &ilf2xy );
CImgl<float> ImgListTrans( CImgl<float> &ilf2a, int iima, int iina );


int readImg( const char* iFile, float* &iBuf )
{
  using namespace std;
  //Open file
  ifstream fF1;
  fF1.open( iFile, ios::in | ios::binary );

  fF1.seekg(0, ios::end);
  int fiSz = fF1.tellg();
  fF1.seekg(0, ios::beg);

  //Create Buffer
  iBuf = new float[fiSz/sizeof(float)];
  //Read contents
  fF1.read( (char*) iBuf, fiSz );

  //Close file
  fF1.close();
  
  return fiSz/sizeof(float);
}

//********************************************************************
int main()
{
  int nlhs = 4;
  //==================================================================
  //Initialisation
  using namespace std;
  
  float *ffp1, *ffp2, *ffp3;
  
  readImg( "ztmp.bin",  ffp1 );
  readImg( "zref.bin",  ffp2 );
  readImg( "zwarp.bin", ffp3 );
  
  //Read the inputs
  CImg<float> ff2tmp ( ffp1, 13,13 ); //Template image
  CImg<float> ff2ref ( ffp2, 160,213 ); //Reference image
  CImg<float> ff2warp( ffp3, 6,1 ); //The Affine warp

  delete[] ffp1;delete[] ffp2;delete[] ffp3;

  //==================================================================
  //Body

  //--------------------------------------------------------------
  //Function
  
  //Extract region in ff2ref corresponding to warped ff2tmp
  CImg<float> ff2region = region(ff2tmp,ff2ref,ff2warp);

  CImg<float> ff2Diff = ff2region - ff2tmp;
  float ofSSD = (ff2Diff).sum();

  //Output result
  //plhs[0] = CImg<float>( 1,1,1,1,ofSSD).toMatlab();
  cout<<"\nofSSD "<<ofSSD;
  if( nlhs <= 1 ) return 0;

  //--------------------------------------------------------------
  //Gradient

  //Extract region from gradient image
  CImgl<float> flf2refGrad = ff2ref.get_gradientXY( 1 );
  CImgl<float> flf2regionGrad( region(ff2tmp,flf2refGrad[0],ff2warp),
			       region(ff2tmp,flf2refGrad[1],ff2warp) );
  //Generate warp derivative components
  CImgl<float> flf2tmpXY = imgXY( ff2tmp );
  //Compute Jacobian Image
  CImgl<float> flf2JacImg( flf2tmpXY[0].get_mul(flf2regionGrad[0]),
			   flf2tmpXY[0].get_mul(flf2regionGrad[1]),
			   flf2tmpXY[1].get_mul(flf2regionGrad[0]),
			   flf2tmpXY[1].get_mul(flf2regionGrad[1]),
			   flf2regionGrad[0], flf2regionGrad[1] );

  //Compute gradient MI
  CImg<float> of2SSDjac(1,6,1,1,0.0);
 
  for( int n=0; n<6; n++ ) {
    of2SSDjac[n] = (flf2JacImg[n].get_mul(ff2Diff) ).sum();
  }
  
  //Output result  
  //plhs[1] = of2SSDjac.toMatlab();
  of2SSDjac.print("\nof2SSDJac ",2);
  if( nlhs <= 2 ) return 0;

  //--------------------------------------------------------------
  //Hessian: 1st order part

  //Compute 1st order Hessian of SSD
  CImg<float> of2SSDhess1(6,6,1,1,0.0);

  for( int m=0; m<6; m++ ) {
    for( int n=0; n<6; n++ ) {
      of2SSDhess1(m,n) = flf2JacImg[n].get_mul(flf2JacImg[m]).sum();
    }}
  //Output result
  //plhs[2] = of2SSDhess1.toMatlab();
  of2SSDhess1.print("\nof2SSDhess1",2);
  if( nlhs <= 3 ) return 0;

  //--------------------------------------------------------------
  //Hessian: 2nd order part

  // (del R dwdv)^2
  CImgl<float> flf2vRdwdvSqd = ImgListMul( flf2JacImg,6,1, flf2JacImg,1,6 );
  // (Del^2 R) (dwdv)^2
  CImgl<float> flf2refDDX = flf2refGrad[0].get_gradientXY(1);
  CImgl<float> flf2refDDY = flf2refGrad[1].get_gradientXY(1);
  CImgl<float> flf2regionDD( region( ff2tmp, flf2refDDX[0], ff2warp ),
			     region( ff2tmp, flf2refDDX[1], ff2warp ),
			     region( ff2tmp, flf2refDDY[0], ff2warp ),
			     region( ff2tmp, flf2refDDY[1], ff2warp ) );

  CImgl<float> flf2dwdv = getdWdV( flf2tmpXY );  
  CImgl<float> flf2dwdvT = ImgListTrans( flf2dwdv,2,6 );

  //Compute
  CImgl<float> flf2tmp = ImgListMul( flf2dwdvT,6,2, flf2regionDD,2,2 );
  CImgl<float> flf2vvRdwdvSqd = ImgListMul( flf2tmp,6,2, flf2dwdv,2,6 );

  CImg<float> of2SSDhess2(6,6,1,1,0.0);
  for( int n=0; n<36; n++ ) {
    of2SSDhess2[n] = (flf2vvRdwdvSqd[n]+flf2vRdwdvSqd[n]).mul(ff2Diff).sum();
  }

  //Output result
  of2SSDhess2.print("\nof2SSDhess2",2);
  //plhs[3] = of2SSDhess2.toMatlab();

  return 0;
}

 
//==================================================================
//Get the joint histogram of two images.
//Assumes the sizes of the 2 images are the same
//Assumes that amax and bmax are correctly sized
CImg<float> jointHist( CImg<float>& if2a, CImg<float>& if2b, 
			int amax, int bmax )
{
  float* fpf1 = if2a.ptr(), *fpf2 = if2b.ptr();
  float* fpfend = fpf1 + if2a.size();
  float ffProb = 1.0/if2a.size();

  //Do we need to assign sizes to the histogram
  if( amax==0 ) amax = (int) CImgStats( if2a ).max + 1;
  if( bmax==0 ) bmax = (int) CImgStats( if2b ).max + 1;

  CImg<float> of2Hist( amax,bmax,1,1, 0.0 );

  //Populate the histogram
  while( fpf1 < fpfend )
    {
      of2Hist( (int) *fpf1, (int) *fpf2 )+=ffProb;
      fpf1++; fpf2++;
    }

  return of2Hist;
}

//==================================================================
//Get the joint histogram of two images - weighting samples by if2val
//Assumes the sizes of the 2 images are the same
//Assumes that amax and bmax are correctly sized
CImg<float> jointHist( CImg<float>& if2a, CImg<float>& if2b,
			CImg<float>& if2val,
			int amax, int bmax )
{
  float *fpf1 = if2a.ptr(), *fpf2 = if2b.ptr(), *fpfval = if2val.ptr();
  float *fpfend = fpf1 + if2a.size();

  //Do we need to assign sizes to the histogram
  if( amax==0 ) amax = (int) CImgStats( if2a ).max + 1;
  if( bmax==0 ) bmax = (int) CImgStats( if2b ).max + 1;

  CImg<float> of2Hist( amax,bmax,1,1, 0.0 );

  //Populate the histogram
  while( fpf1 < fpfend )
    {
      of2Hist( (int) *fpf1, (int) *fpf2 ) += *fpfval;
      fpf1++; fpf2++; fpfval++;
    }
  of2Hist /= if2a.size();

  return of2Hist;
}

//==================================================================
//Sums a given image along a particular dimension
CImg<float> sumDim( CImg<float>& if2img, int iiDim )
{
  //Which dimension are we summing along
  int fid[4], fids[4];
  fid[0] = if2img.dimx(); fid[1] = if2img.dimy();
  fid[2] = if2img.dimz(); fid[3] = if2img.dimv();
  for( int i=0; i<4; i++ ) fids[i] = fid[i];
  if( iiDim < 1 ) iiDim=1;
  if( iiDim > 4 ) iiDim=1;
  iiDim--;
  fids[iiDim] = 1;

  float* fpfimg = if2img.ptr(); //Pointer to img

  CImg<float> of2ImgSum( fids[0], fids[1], fids[2], fids[3], 0.0 );

  for( int fiv=0; fiv<if2img.dimv(); fiv++ ) 
    { 
      int fiv2 = iiDim==3 ? 0 : fiv;
      for( int fiz=0; fiz<if2img.dimz(); fiz++ ) 
	{
	  int fiz2 = iiDim==2 ? 0 : fiz;
	  for( int fiy=0; fiy<if2img.dimy(); fiy++ ) 
	    {
	      int fiy2 = iiDim==1 ? 0 : fiy;
	      for( int fix=0; fix<if2img.dimx(); fix++ )
		{
		  int fix2 = iiDim==0 ? 0 : fix;
		  of2ImgSum( fix2, fiy2, fiz2, fiv2 ) += *fpfimg;
		  fpfimg++;
		}}}}

  return of2ImgSum;
}

//==================================================================
//Measure the entropy of a given histogram.
//Assumes everything sums to one.
float entropy( CImg<float>& if2hist )
{
  float *fpfPixel = if2hist.ptr(), *fpfEnd = fpfPixel+if2hist.size();
  float ofEntropy = 0.0;

  while( fpfPixel < fpfEnd )
    {
      float v = *fpfPixel;

      if( v>0.0 ) 
	ofEntropy += v*log(v);

      fpfPixel++;
    }
  return -ofEntropy;
}

//==================================================================
//Sums a given image along a particular dimension
//Assumes images are 2D - does linear interpolation
CImg<float> region( CImg<float>& if2tmp, CImg<float>& if2ref,  
		    CImg<float>& if2warp )
{
  CImg<float> of2region( if2tmp, false );
  of2region=0.0;
  float *fpfPixel = of2region.ptr();

  int fiyA = -(if2tmp.dimy()-1)/2, fiyE = (if2tmp.dimy()-1)/2;
  int fixA = -(if2tmp.dimx()-1)/2, fixE = (if2tmp.dimx()-1)/2;

  for( int fiy=fiyA; fiy<=fiyE; fiy++ )
    for( int fix=fixA; fix<=fixE; fix++ )
      {
	float ffxw = (1+if2warp[0])*fix +    if2warp[2] *fiy + if2warp[4];
	float ffyw =    if2warp[1] *fix + (1+if2warp[3])*fiy + if2warp[5];
	(*fpfPixel) = if2ref.linear_pix2d(ffxw,ffyw);

//  	std::cout<<"\ntx,y rx,y v: "<< fix<<","<<fiy<<" "
//  		 <<ffxw<<","<<ffyw<<" "<<(*fpfPixel);
	fpfPixel++;
      }

  return of2region;
}

//==================================================================
//Sums a given image along a particular dimension
//Assumes images are 2D - does linear interpolation
CImgl<float> imgXY( CImg<float>& if2tmp )
{
  CImgl<float> olf2XY( 2, if2tmp.dimx(), if2tmp.dimy() ); 
//   olf2XY[0]=0.0;
//   olf2XY[1]=0.0;

  float half = olf2XY[0].dimx()/2;
  cimg_mapXY(olf2XY[0],x,y) olf2XY[0](x,y)=x-half;

  half = olf2XY[0].dimy()/2;
  cimg_mapXY(olf2XY[1],x,y) olf2XY[1](x,y)=y-half;
  return olf2XY;
}

//==================================================================
//Output an image in matrix form
void outputImg( CImg<float> &if2a, char* iName )
{
  using namespace std;
  cout<<"\n"<<iName<<" = reshape([";
  for( int i=0; i<if2a.size(); i++ ) cout<<if2a[i]<<" ";
  cout<<"],["<<if2a.dimx()<<" "<<if2a.dimy()<<"]);\n";
}

//==================================================================
//Multiply two matrices of images together
//Assumes images are all same size and that iina==iimb
CImgl<float> ImgListMul( CImgl<float> &ilf2a, int iima, int iina,
			 CImgl<float> &ilf2b, int iimb, int iinb )
{
  CImgl<float> olf2Result( iima*iinb, ilf2a[0].dimx(), ilf2a[0].dimy(),
			   ilf2a[0].dimz(), ilf2a[0].dimv(), 0.0 );
  
  for( int fim=0; fim<iima; fim++ ) {
    for( int fin=0; fin<iinb; fin++ ) {
      for( int fie=0; fie<iina; fie++ ) 
	{
	  //olf2Result(fim,fin) += ilf2a(fim,fie) * ilf2b(fie,fin);
	  olf2Result[fim+fin*iima] += //iima*iinb
	    ilf2a[fim+fie*iima].get_mul( //iima*iina
	    ilf2b[fie+fin*iimb] );  //iimb*iinb
	}}}      

  return olf2Result;
}

//==================================================================
//Multiply two matrices of images together
//Assumes images are all same size and that iina==iimb
CImgl<float> getdWdV( CImgl<float> &ilf2xy )
{
  CImgl<float> olf2dWdV( 12, ilf2xy[0].dimx(), ilf2xy[0].dimy(), 
			 ilf2xy[0].dimz(), ilf2xy[0].dimv() );

  olf2dWdV[ 0] = ilf2xy[0]; olf2dWdV[ 1] = 0.0 ;
  olf2dWdV[ 2] = 0.0;       olf2dWdV[ 3] = ilf2xy[0];
  olf2dWdV[ 4] = ilf2xy[1]; olf2dWdV[ 5] = 0.0 ;
  olf2dWdV[ 6] = 0.0;       olf2dWdV[ 7] = ilf2xy[1];
  olf2dWdV[ 8] = 1.0;       olf2dWdV[ 9] = 0.0 ;
  olf2dWdV[10] = 0.0;       olf2dWdV[11] = 1.0 ;
  
  return olf2dWdV;
}

//==================================================================
//Transpose a matrix of images
//Assumes images are all same size
CImgl<float> ImgListTrans( CImgl<float> &ilf2a, int iima, int iina )
{
  CImgl<float> olf2trans( ilf2a );
  
  for( int fim=0; fim<iina; fim++ ){
    for( int fin=0; fin<iima; fin++ )
      {
	//olf2trans(fim,fin) = ilf2a(fin,fim);
	olf2trans[fim+fin*iina] = //iina*iima
	  ilf2a[fin+fim*iima]; //iima*iina
      }}

  return olf2trans;
}
